unit FixObj;

interface
uses
  sysutils;

Type
  // This enumeration is a list of all the object types in the HP49
  // Note: the order must be the same that the PrologsNames const.
  // wgg keep HP49 specific items at the end so we can prune them for HP48 use
  TPrologs =( DOACPTR  ,
              DOARRY   , //strobj
              DOBAK    , //strobj
              DOBINT   ,
              DOCHAR   ,
              DOCMP    ,
              DOCODE   , //strobj
              DOCOL    ,
              DOCSTR   , //strobj
              DOECMP   , //strobj
              DOEREL   , //strobj
              DOEXT    ,
              DOEXT0   ,
              DOEXT2   , //strobj
              DOEXT3   , //strobj
              DOEXT4   , //strobj
              DOGROB   , //strobj
              DOHXS    , //strobj
              DOIDNT   ,
              DOLAM    ,
              DOLIB    , //strobj
              DOLIST   ,
              DOLNKARRY, //strobj
              DOREAL   ,
              DOROMP   ,
              DORRP    ,
              DOSYMB   ,
              DOTAG    ,
              SEMI     ,
              // wgg HP49 specific items
              DOAPLET  , //strobj
              DOFLASHP ,
              DOINT    ,
              DOLNGCMP ,
              DOLNGREAL,
              DOMATRIX ,
              DOMINIFONT );  //strobj

  // This type intent to save the value of every prolog in the HP49.
  TPrologsList = array [TPrologs] of integer;

  // this function allow the decompilation of the object at a specified address
  // adr: the adr of the object to decompile (at the end, it's the adr of the end of the object)
  // the 2 next parameters intent to receive a function alloging to read from the memory
  // Prolog intent to receive a list of the value of each prolog
  // the function returns a string representing the object

  procedure DecompObj(Var adr: integer);
  function FixFile(var s: string): boolean;                     // wgg file to set length

implementation
uses
  utils;

const
  // This is the list of the prolog constants. Values are used to retrive the
  // numerical value from the .lr files.
  // Note: the order must be the same that the TPrologs type.
  Prolog : array [TPrologs] of integer =(
    $02BAA,
    $029E8, 
    $02B62,
    $02911,
    $029BF,
    $02977,
    $02DCC,
    $02D9D,
    $02A2C,
    $0299D,
    $02955,
    $02ADA,
    $02B88,
    $02BCC,
    $02BEE,
    $02C10,
    $02B1E,
    $02A4E,
    $02E48,
    $02E6D,
    $02B40,
    $02A74,
    $02A0A,
    $02933,
    $02E92,
    $02A96,
    $02AB8,
    $02AFC,
    $0312B,
  // wgg HP49 specific items, if this list changes patch InitPrologues below
    $026D5,
    $026AC,
    $02614,
    $02660,
    $0263A,
    $02686,
    $026FE  );


var
  HPobj: string;
  pos: integer;             // wgg position in nibbles 1...
  Is49: boolean;
  MaxProlog: TPrologs;
  DecompResult: boolean;

// wgg read an HPHP4x file and adjust its length using the Decom[p routines
function FixFile(var s:string): boolean;
begin
  HPobj:= s;
  MaxProlog:= High(Prolog);
  DecompResult:= true;
  pos:= 16;
  if HPobj[6]='9' then Is49:= true
  else
  begin
    Is49:= false;
    MaxProlog:= SEMI;
  end;
  DecompObj(pos);
  setlength(HPobj, (pos+1) div 2);     // in case of odd nibble
  s:= HPobj;
  result:= DecompResult;
end;

// wgg get an HP nibble starting at nibble, adr
// nibbles are numbered starting with 0...
function Nibble(adr: integer): integer;
begin
  result:= ord(HPobj[(adr+2) div 2]);
  if (adr And 1) = 0 then result:= (result And $0F)  // low nibble is first
  else result:= (result And $F0) shr 4;
end;

// wgg get an HP integer starting at nibble, adr
function ReadInteger(adr: integer): integer;
begin
  result:=nibble(adr);
  result:= result+nibble(adr+1)*16;
  result:= result+nibble(adr+2)*256;
  result:= result+nibble(adr+3)*4096;
  result:= result+nibble(adr+4)*65536;
end;

// wgg get an HP integer starting at nibble, adr of length N
function ReadIntN(adr: integer; N: integer): integer;
begin
  result:= 0;
  while N > 0 do
  begin
    dec(N);
    result:= result*16+nibble(adr+N);
  end;
end;

// This is the main function doing the work.
procedure DecompObj(Var adr: integer);
var
  time: TDateTime;

  procedure DecompObj2(var adr: integer); forward;
    // array decompilation (to be enhanced)
    Function ARRY(adr: integer): integer;
    Begin
      result:= adr+5+ReadInteger(adr+5);
    end;
    // binary integer decompilation
    Function BINT(adr: integer): integer;
    Begin
      result:= adr+10;
    end;
    // character decompilation (to be enhanced)
    Function CHAR(adr: integer): integer;
    Begin
      result:= adr+7;
    end;
    // normal complex decompilation (to be enhanced)
    Function CMP (adr: integer): integer;
    Begin
      result:= adr+5+16+16;
    end;
    // Assembly program decompilation (Need to add Hex)
    Function CODE(adr: integer): integer;
    Begin
      result:= adr+5+ReadInteger(adr+5);
    end;
    // Rpl program decompilation
    Function COL (adr: integer): integer;
    Begin
      inc(adr, 5);
      While ReadInteger(adr) <> Prolog[SEMI] Do
        DecompObj2(adr);
      result:= adr+5;
    end;
    // chr string decompilation (Need to be speeded up for long strings)
    Function CSTR(adr: integer): integer;
    Begin
      result:= adr+ ReadInteger(adr+5)+5;
    end;
    // Units decompilation
    Function EXT (adr: integer): integer;
    Begin
      adr := adr + 5;
      While ReadInteger(adr) <> Prolog[SEMI] Do
        DecompObj2(adr);
      result:= adr+5;
    end;
    // library datas (hexa?)
    Function EXT0(adr: integer): integer;
    Begin
      result:= adr+ ReadInteger(adr+5)+5;
    end;
    // Extended pointer. (Try to resolve the extended pointer if in ROM)
    Function ACPTR(adr: integer): integer;
    Begin
      result := adr + 15;
    end;
    // Extended objects. (Hexa?)
    Function EXT2(adr: integer): integer;
    Begin
      result:= adr+ ReadInteger(adr+5)+5;
    end;
    Function EXT3(adr: integer): integer;
    Begin
      result:= adr+ ReadInteger(adr+5)+5;
    end;
    Function EXT4(adr: integer): integer;
    Begin
      result:= adr+ ReadInteger(adr+5)+5;
    end;
    // Graphic. (Need a graphic editor)
    Function GROB(adr: integer): integer;
    Begin
      result:= adr+ReadInteger(adr+5)+5;
    end;
    // binary integer
    Function HXS(adr: integer): integer;
    Begin
      result:= adr+ReadInteger(adr+5)+5;
    end;
    // Identifier decompilation (need to avoid stranges chr like in
    // DOCSTR need to create a ReadChr function)
    Function IDNT(adr: integer): integer;
    var
      i: integer;
    Begin
      i:= ReadInteger(adr+5) and $FF;
      result:= adr+i*2+7
    end;
    // Local Identifier decompilation (need to avoid stranges chr like in DOCSTR need to create a ReadChr function)
    Function LAM (adr: integer): integer;
    var
      i: integer;
     Begin
      i:= ReadInteger(adr+5) and $FF;
      result:= adr+i*2+7
    end;
    // Library decompilation (need to put either the name and number or an hexa rep)
    Function LIB (adr: integer): integer;
    Begin
      result:= adr+ReadInteger(adr+5)+5
    end;
    // Backup (need to decompile)
    Function BAK (adr: integer): integer;
    Begin
      result:= adr+ReadInteger(adr+5)+5
    end;
    // List decompilation
    Function LIST(adr: integer): integer;
    Begin
      inc(adr, 5);
      While ReadInteger(adr) <> prolog[SEMI] Do
        DecompObj2(adr);
      result:= adr+5;
    end;
    // real decompilation. (need to improve the decompilation)
    Function REAL(adr: integer): integer;
    Begin
      result:= adr+5+16;
    end;
    // Extended real. (Need real decompilation)
    Function EREL(adr: integer): integer;
    Begin
      result:= adr+5+16+5;
    end;
    // Extended complex. (Need real decompilation)
    Function ECMP(adr: integer): integer;
    Begin
      result:= adr+5+16+5+5+16;
    end;
    // Rompointer. (We should try to get the real value)
    Function ROMP(adr: integer): integer;
    Begin
      result := adr + 5 + 6;
    end;
    // Linked array. (need real decompilation)
    Function LNKARRY(adr: integer): integer;
    Begin
      result:= adr+ReadInteger(adr+5)+5
    end;
    // symbolic object decompilation
    Function SYMB(adr: integer): integer;
    Begin
      inc(adr, 5);
      While ReadInteger(adr) <> prolog[SEMI] Do
        DecompObj2(adr);
      result:= adr+5;
    end;
    // tag object decompilation. (Need to use a getChr function)
    Function TAG(adr: integer): integer;
    var
      i: integer;
    Begin
      i:= ReadInteger(adr+5) and $FF;
      result:= adr+i*2+7;
      DecompObj2(result);
    end;
    // Long integer decompilation. No syntax yet
    Function INT(adr: integer): integer;
    var
      i: integer;
    Begin
      i:= ReadInteger(adr+5);
      result:= adr+i+5
    end;
    // new Long real decompilation. No syntax yet
    Function LNGREAL(adr: integer): integer;
    Begin
      adr:= INT(adr);
      result:= INT(adr-5);
    end;
    // new Long complex decompilation. No syntax yet
    Function LNGCMP(adr: integer): integer;
    Begin
      adr:= LNGREAL(adr);
      result:= LNGREAL(adr-5);
    end;
    // flashpointer. (Need to put the symbol of exists)
    Function FLASHP(adr: integer): integer;
    Begin
      result:= adr+5+3+4;
    end;
    // new symbolic matrix decompilation.
    Function MATRIX(adr: integer): integer;
    Begin
      inc(adr, 5);
      While ReadInteger(adr) <> Prolog[SEMI] Do
        DecompObj2(adr);
      result:= adr+5;
    end;

    // wgg directory
    Function RRP(adr: integer): integer;
    Begin
      adr:= ReadInteger(adr+8) + adr + 8 ;  // points to count field of last name
      adr:= adr + ReadIntN(adr, 2)*2 + 4;   // point at last object,
      DecompObj2(adr);                      //    we used + 4 because its N string N
      result:= adr;
    end;

  procedure DecompObj2(var adr: integer);
  Var
    i: integer;
    j : TPrologs;
  Begin
    Try
      if now-time > 1/24/3600*3 then
        raise exception.create('TimeOut');
      i:= ReadInteger(adr);                    // read the data at the address to decompile
      j:= Low(Prolog);                         // search for it in the prolog list
      while (j<=MaxProlog) and (prolog[j]<>i) do inc(j);
      // if it is a prolog, we are pointing directly on the object, decompile it
      case j of
        DOACPTR    : adr := ACPTR   (adr);
        DOARRY     : adr := ARRY    (adr);
        DOBAK      : adr := BAK     (adr);
        DOBINT     : adr := BINT    (adr);
        DOCHAR     : adr := CHAR    (adr);
        DOCMP      : adr := CMP     (adr);
        DOCODE     : adr := CODE    (adr);
        DOCOL      : adr := COL     (adr);
        DOCSTR     : adr := CSTR    (adr);
        DOECMP     : adr := ECMP    (adr);
        DOEREL     : adr := EREL    (adr);
        DOEXT      : adr := EXT     (adr);
        DOEXT0     : adr := EXT0    (adr);
        DOEXT2     : adr := EXT2    (adr);
        DOEXT3     : adr := EXT3    (adr);
        DOEXT4     : adr := EXT4    (adr);
        DOHXS      : adr := HXS     (adr);
        DOIDNT     : adr := IDNT    (adr);
        DOLAM      : adr := LAM     (adr);
        DOLIB      : adr := LIB     (adr);
        DOLIST     : adr := LIST    (adr);
        DOLNKARRY  : adr := LNKARRY (adr);
        DOREAL     : adr := REAL    (adr);
        DOROMP     : adr := ROMP    (adr);
        DOSYMB     : adr := SYMB    (adr);
        DOTAG      : adr := TAG     (adr);
        DORRP      : adr := RRP     (adr);
        DOGROB     : adr := GROB    (adr);
//      DOFONT     : adr := FONT    (adr);       // unknown, not implemented
        // HP49 specific items
//      DOAPLET    : adr := APLET   (adr);       // not implemented
        DOFLASHP   : adr := FLASHP  (adr);
        DOINT      : adr := INT     (adr);
        DOLNGCMP   : adr := LNGCMP  (adr);
        DOLNGREAL  : adr := LNGREAL (adr);
        DOMATRIX   : adr := MATRIX  (adr);
//      DOMINIFONT : adr := MINIFONT(adr);       // not implemented

        // it is not a prolog.
        // if it is a pointer on a simple object, decompile this object,
        // else, if the pointer has a name, give the name
        // else, give the value of the pointer.
        else
          inc(adr,5);          // assume just a ptr
      end;
    Except
    End;
  End;

Begin
  time:= now;
  DecompObj2(adr);
end;

end.
